#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
class AggregationExecutor : public AbstractExecutor {
private:
    std::unique_ptr<AbstractExecutor> prev_;        // 聚集函数的投影儿子节点
    std::vector<ColMeta> agre_cols_;                     // 需要计算的字段
    std::vector<ColMeta> res_;                      // 输出的字段
    std::vector<AgreValue> agre_values_;            // 计算的结果
    std::vector<AgreCol> agre_cols_meta_;              //聚合列表的元数据集
    std::vector<TabCol> groupby_cols_;              //groupby列
    std::vector<AgreCondition> having_conds_;        //having条件
    size_t len_;                                    // 聚集后结果的字段总长度

    std::map<std::string, std::vector<AgreValue>> grouped_values_; // 用于保存按group by分组后的聚合结果
    std::map<std::string, std::vector<AgreValue>>::iterator group_iter_; // 迭代器用于遍历分组结果

    std::map<std::string, std::vector<AgreValue>> grouped_having_conds;

    bool only_count = false;                         //一种特殊情况，只有count聚合类型,且无需理会上层传来的具体数据
    bool end_flag = false;

public:
    AggregationExecutor(std::unique_ptr<AbstractExecutor> &prev,
                        std::vector<AgreCol> agre_col_meta,
                        std::vector<TabCol> groupby_cols,
                        std::vector<AgreCondition> having_conds) {
        prev_ = std::move(prev);//得到所有的列
        having_conds_ = std::move(having_conds);
        groupby_cols_ = std::move(groupby_cols);
        agre_cols_meta_ = std::move(agre_col_meta);

        only_count = (groupby_cols_.empty())&&((agre_cols_meta_.size() == 1) && ((agre_cols_meta_[0].agre_type == AGRE_TYPE_COUNT_ALL) || (agre_cols_meta_[0].agre_type == AGRE_TYPE_COUNT)));
        //初始化res_和cols
        init_resCol_and_agreCol();
    }

    void beginTuple() override {

        prev_->beginTuple();
        //case:where筛选导致上层没有返回数据
        if (prev_->is_end()) {
            // 初始化为空结果
            grouped_values_[""] = init_aggregation_values();
            group_iter_ = grouped_values_.begin();
            return;
        }
        std::string group_key = "";
        auto &prev_cols = prev_->cols();
        std::unique_ptr<RmRecord> rec_buffer;
        while (!prev_->is_end()) {
            if(deal_with_only_count())continue;//无需具体数据
            rec_buffer = prev_->Next();
            if(!groupby_cols_.empty()){
                group_key = generate_group_key(rec_buffer);
                if (grouped_having_conds.find(group_key) == grouped_having_conds.end()) {
                    // 初始化新的分组
                    grouped_having_conds[group_key] = init_having_clause_values();
                }
                update_aggregation_values(grouped_having_conds[group_key],rec_buffer,having_conds_);
            }
            if (grouped_values_.find(group_key) == grouped_values_.end()) {
                // 初始化新的分组
                grouped_values_[group_key] = init_aggregation_values();
            }
            // 更新聚合值
            auto &values = grouped_values_[group_key];
            for (auto val_iter = values.begin(); val_iter != values.end(); ++val_iter) {
                if(val_iter->agre_type == AgreType::AGRE_TYPE_COUNT_ALL){
                    val_iter->add_value(nullptr,sizeof(int));
                }else{
                    auto col_meta = get_col(prev_cols,agre_cols_[val_iter - values.begin()]);
                    val_iter->add_value(rec_buffer->data + col_meta->offset, col_meta->len);
                }
            }
            prev_->nextTuple();
        }
        filter_having_conditions();
        group_iter_ = grouped_values_.begin();
        if(grouped_values_.empty())end_flag = true;
    }

    void nextTuple() override {
        if (group_iter_ != grouped_values_.end()) {
            ++group_iter_;
        }

        if (group_iter_ == grouped_values_.end()) {
            end_flag = true;
        }
    }

    bool is_end() const override {
        return end_flag;
    }


    std::unique_ptr<RmRecord> Next() override {
        if (is_end()) {
            return nullptr;
        }
        auto &group_key = group_iter_->first;
        auto &current_group = group_iter_->second;
        std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>(this->len_);
        int curr_offset = 0;
        int key_offset = 0;
        for (auto &col : groupby_cols_) {
            auto col_meta = get_col_meta(prev_->cols(), col);
            int col_len = col_meta->len;
            memcpy(ret->data + curr_offset, group_key.data() + key_offset, col_len);
            curr_offset += col_len;
            key_offset += col_len;
        }

        // 将聚合值写入结果记录
        for (auto &value : current_group) {
            memcpy(ret->data + curr_offset, value.get_rec()->data, value.get_len());
            curr_offset += value.get_len();
        }
        return ret;
    }

    Rid &rid() override { return _abstract_rid; }

    const std::vector<ColMeta> &cols() const {
        return this->res_;
    };

    size_t tupleLen() const {
        return this->len_;
    }

    std::string getType() { return "AggreagtionExecutor"; };

    /**
     *
     * @param col_metas 上层列元数据
     * @param target  目标列，TabCol形式
     * @return target的列元数据
     * @brief 这里忽略了taregt的表名。由于表连接，这里可能会有两个相同的target。需要确保这里仅仅用于groupby列的选取
     */
    std::vector<ColMeta>::const_iterator get_col_meta(const std::vector<ColMeta>& col_metas, const TabCol &target) {
        auto pos = std::find_if(col_metas.begin(), col_metas.end(), [&](const ColMeta &col) {
            return col.name == target.col_name;
        });
        if (pos == col_metas.end()) {
            throw ColumnNotFoundError(target.tab_name + '.' + target.col_name);
        }
        return pos;
    }

    ColMeta get_col_offset(const TabCol &target) override{
        for(auto iter = this->res_.begin(); iter!=this->res_.end(); iter++){
            if(iter->tab_name == target.tab_name && iter->name == target.col_name){
                return *iter;
            }
        }
        return ColMeta();
    }

    void init_resCol_and_agreCol(){
        size_t res_offset = 0;
        auto &prev_cols = prev_->cols();
        for(auto &[_,col_name] : groupby_cols_){
            auto col_iter = std::find_if(prev_cols.begin(),prev_cols.end(),[&](const ColMeta &col){
                return (col_name == col.name);
            });
            //assert(col_iter != prev_cols.end());
            auto regular_col = *col_iter;
            regular_col.offset = res_offset;
            res_offset += regular_col.len;
            regular_col.is_agrregated = false;
            res_.emplace_back(regular_col);
        }
        for (auto &[alias,agre_col,agre_type] : agre_cols_meta_) {
            //维护输出结果
            ColMeta res_col;
            switch (agre_type) {
                case AGRE_TYPE_COUNT_ALL:
                //case AGRE_TYPE_COUNT:
                    res_col.type = TYPE_INT;
                    res_col.len = sizeof(int);
                    //res_col.tab_name = col.ta
                    break;
                default:
                    //其他聚合列
                    auto col = *std::find_if(prev_cols.begin(),prev_cols.end(),[&](const ColMeta &col){
                        return agre_col.col_name == col.name;
                    });
                    res_col.len = col.len;
                    res_col.type = col.type;
                    res_col.tab_name = col.tab_name;
                    res_col.name = col.name;
            }

            res_col.offset = res_offset;
            res_offset += res_col.len;
            res_col.is_agrregated = true;
            agre_cols_.emplace_back(res_col);

            //维护表头
            res_col.name = alias.col_name;
            if(agre_type == AGRE_TYPE_COUNT){
                res_offset = res_offset - res_col.len + sizeof(int);
                //res_col.offset = res_offset;
                res_col.len = sizeof(int);
                res_col.type = TYPE_INT;
            }
            res_.emplace_back(res_col);//表头
        }
        len_ = res_offset;
    }


    bool deal_with_only_count(){
        if(!only_count)return false;
        auto group_key = "";//保证group_col为0
        if (grouped_values_.find(group_key) == grouped_values_.end()) {
            // 初始化新的分组
            grouped_values_[group_key] = init_aggregation_values();
        }
        grouped_values_[group_key][0].int_val++;
        //grouped_values_[group_key][0].add_value(nullptr,sizeof(int));
        prev_->nextTuple();
        return true;
    }
    std::string generate_group_key(const std::unique_ptr<RmRecord> &tuple) {
        std::string key;
        for (const auto &col : groupby_cols_) {
            auto col_meta = get_col_meta(prev_->cols(), col);
            key.append(tuple->data + col_meta->offset, col_meta->len);
        }
        return key;
    }

    /**
     *
     * @return
     */
    std::vector<AgreValue> init_aggregation_values() {
        std::vector<AgreValue> values;
        for (int i = 0; i < (int)agre_cols_meta_.size();i++) {
            AgreValue value;
            // 初始化聚合值
            auto agre_type = agre_cols_meta_[i].agre_type;
            if(agre_type == AGRE_TYPE_COUNT_ALL){
                value.init_agre_raw(TYPE_INT, AGRE_TYPE_COUNT_ALL, sizeof(int));
//            }else if(agre_type == AGRE_TYPE_COUNT){
//                value.init_agre_raw(TYPE_INT, AGRE_TYPE_COUNT, sizeof(int));
            }else{
                value.init_agre_raw(agre_cols_[i].type, agre_cols_meta_[i].agre_type,agre_cols_[i].len);
            }
            values.emplace_back(value);
        }
        return values;
    }

    std::vector<AgreValue> init_having_clause_values(){
        std::vector<AgreValue> values;
        for (int i = 0; i < (int)having_conds_.size();i++) {
            AgreValue value;
            auto agre_type = having_conds_[i].agre_lhs_col.agre_type;
            auto &left_table_col = having_conds_[i].agre_lhs_col.agre_col;//特殊情况,left table没有名字，此时应该是count*
            auto &prev_cols = prev_->cols();
            // 初始化聚合值
            if(agre_type == AGRE_TYPE_COUNT_ALL){
                value.init_agre_raw(TYPE_INT, AGRE_TYPE_COUNT_ALL, sizeof(int));
//            }else if(agre_type == AGRE_TYPE_COUNT){
//                value.init_agre_raw(TYPE_INT, AGRE_TYPE_COUNT, sizeof(int));
            }else{
                auto col_iter = get_col(prev_cols,left_table_col);
                value.init_agre_raw(col_iter->type,agre_type,col_iter->len);//可能有错
            }
            values.emplace_back(value);
        }
        return values;
    }
    /**
     *
     * @param values 已经初始化的，存放新的聚合值
     * @param tuple  上层得到的按物理地址排序的fields
     * @param agre_cols 目标聚合列
     */
    void update_aggregation_values(std::vector<AgreValue> &values,
                                   const std::unique_ptr<RmRecord> &tuple,
                                   std::vector<ColMeta> &agre_cols) {
        auto &prev_cols = prev_->cols();
        for (size_t i = 0; i < values.size(); ++i) {
            if(values[i].agre_type == AgreType::AGRE_TYPE_COUNT_ALL){
                values[i].add_value("", sizeof(int));
            }else{
                auto col_meta = get_col(prev_cols,agre_cols[i]);
                char *value_buf = tuple->data + col_meta->offset;//
                values[i].add_value(value_buf, col_meta->len);
            }
        }
    }

    void update_aggregation_values(std::vector<AgreValue> &values,
                                   const std::unique_ptr<RmRecord> &tuple,
                                   std::vector<AgreCondition> &agre_conditions) {
        auto &prev_cols = prev_->cols();
        for (size_t i = 0; i < values.size(); ++i) {
            if(values[i].agre_type == AgreType::AGRE_TYPE_COUNT_ALL){
                values[i].add_value("", sizeof(int));
                continue;
            }
            auto col_meta = get_col(prev_cols,agre_conditions[i].agre_lhs_col.agre_col);//添加&
            char *value_buf = tuple->data + col_meta->offset;//
            values[i].add_value(value_buf, col_meta->len);
        }
    }

    /**
     * 用于groupby存在时，having条件
     */
    void filter_having_conditions() {
        for (auto it = grouped_values_.begin(); it != grouped_values_.end(); ) {
            auto group_key = it->first;
            if (!evaluate_having_conditions(grouped_having_conds[group_key])) {
                it = grouped_values_.erase(it);
            } else {
                ++it;
            }
        }
    }

    /**
     * @brief 保证having_conds和values内的次序相同,value为max(id) < 3中的 3
     */
    bool evaluate_having_conditions(const std::vector<AgreValue> &values) {
        for (int i = 0; i < (int)having_conds_.size();i++) {
            const auto lhs_value = values[i].to_value();
            auto having_cond = having_conds_[i];
            if(!compare(having_cond,lhs_value))return false;
        }
        return true;
    }

    bool compare(const AgreCondition &agreCondition,const Value &lhs_value){
        switch (agreCondition.op) {
            case OP_EQ:
                return lhs_value == agreCondition.rhs_val;
            case OP_NE:
                return lhs_value != agreCondition.rhs_val;
            case OP_LT:
                return lhs_value < agreCondition.rhs_val;
            case OP_LE:
                return !(lhs_value > agreCondition.rhs_val);
            case OP_GT:
                return lhs_value > agreCondition.rhs_val;
            case OP_GE:
                return (lhs_value < agreCondition.rhs_val);
            default:
                break;
        }
        return true;
    }
};
