/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include "set"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    std::vector<Condition> conds_;              // 按照index字段顺序排序好的扫描条件 ,最后只剩下非index相关的条件
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    //std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同


    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<RecScan> scan_;
    //RmPageHandle page_buffer;
    SmManager *sm_manager_;
    IxIndexHandle* ih_;
    bool isend;
    bool is_only_index_cond;
    Iid left_;
    Iid right_;
    Iid gap_left_;
    Iid gap_right_;
    bool is_equal;
    std::string index_name_;
    std::unique_ptr<RmRecord> record_buffer;

public:
    IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string>& index_col_names,Context *context)
        :tab_name_(std::move(tab_name)),
         tab_(sm_manager->db_.get_table(tab_name_)),
         conds_(std::move(conds)),
         fh_(sm_manager->fhs_.at(tab_name_).get()),
         cols_(tab_.cols),
         index_meta_(*(tab_.get_index_meta(index_col_names))),
         //page_buffer(&(fh_->get_file_hdr()),(Page*)nullptr),
         sm_manager_(sm_manager)
    {
        context_ = context;
        len_ = cols_.back().offset + cols_.back().len;

        is_equal = false;
        isend = false;
        index_name_ = sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta_.cols);
        ih_ = sm_manager_->ihs_.at(index_name_).get();

        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());//加大锁版本
//        if(conds_.empty()){
//            context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
//        }else{
//            context_->lock_mgr_->lock_IS_on_table(context_->txn_, fh_->GetFd());
//        }
    }

    ~IndexScanExecutor() override {
//        if(page_buffer.page != nullptr){
//            sm_manager_->get_bpm()->unpin_page({fh_->GetFd(),page_buffer.page->get_page_id().page_no}, false);
//        }
    }

    void beginTuple() override {
        GetBound();
        scan_ = std::make_unique<IxScan>(ih_, left_, right_, sm_manager_->get_bpm());
//        if(!is_equal){
//            if(context_->txn_->get_isolation_level() == IsolationLevel::SERIALIZABLE){
//                ih_->gap_lock(gap_left_,gap_right_,context_,fh_->GetFd());
//            }
//        }
        while(!scan_->is_end()) {
            // 遍历
            rid_ = scan_->rid();
            record_buffer = fh_->get_record(rid_, nullptr);
            //record_buffer = fh_->get_record(scan_->rid(),page_buffer, nullptr);//或许加锁
            //可能存在冗余检查，因为之前索引部分的条件依据通过过滤得到
            //bool check = check_all_conditions(record_buffer);
            if (check_all_conditions(record_buffer)) {
                rid_ = scan_->rid();
                break;
            }
            scan_->next();
        }
    }

    void nextTuple() override {
        scan_->next();
        //cond不为空
        while (!scan_->is_end()) {
            record_buffer = fh_->get_record(scan_->rid(), nullptr);
            //record_buffer = fh_->get_record(scan_->rid(),page_buffer, nullptr);
            //bool check = check_all_conditions(record_buffer);
            if (check_all_conditions(record_buffer)) {
                rid_ = scan_->rid();
                break;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        //note:需要注意，其他地方反复调用同一个next，会返回nullptr，如nestloop
        return std::move(record_buffer);
    }

    size_t tupleLen() const override { return len_; }

    std::string getType() override { return "IndexScanExecutor"; }

    const std::vector<ColMeta> &cols() const override { return cols_; }

    ColMeta get_col_offset(const TabCol &target) override {
        ColMeta result;
        for (const auto &col : cols_) {
            if (col.tab_name == target.tab_name && col.name == target.col_name) {
                result = col;
                break;
            }
        }
        return result;
    }
    bool is_end() const override {
        return scan_->is_end();
    }

    Rid &rid() override { return rid_; }

    void GetBound() {
        char *lower = new char[index_meta_.col_tot_len];
        char *upper = new char[index_meta_.col_tot_len];
        bool set_lower = false, set_upper = false;

        // 遍历索引中所有的列，将lower和upper填满
        int current_index = 0;
        char *cur_max;
        char *cur_min;
        for (auto &index_col : index_meta_.cols) {
            int len = index_col.len;
            auto col_name = index_col.name;
            cur_max = get_type_max(index_col);
            cur_min = get_type_min(index_col);
            char *d1 = cur_max;
            char *d2 = cur_min;
            for (const auto &cond : conds_) {
                if(!cond.is_rhs_val)continue;
                if (cond.lhs_col.col_name == col_name) {
                    //假设此时cond内不包含其他表的信息
                    // 找到了条件中的指定列
                    char *right_val_ptr = cond.rhs_val.raw->data;
                    if (cond.op == OP_EQ) {
                        if(ix_compare(cur_min,right_val_ptr,index_col.type,len) < 0){
                            cur_min = right_val_ptr;
                            set_lower = true;
                        }
                        if(ix_compare(cur_max,right_val_ptr,index_col.type,len) > 0){
                            cur_max = right_val_ptr;
                            set_upper = true;
                        }
                    } else if (cond.op == OP_GE || cond.op == OP_GT) {
                        if(ix_compare(cur_min,right_val_ptr,index_col.type,len) < 0){
                            cur_min = right_val_ptr;
                            set_lower = true;
                        }
                    } else if (cond.op == OP_LE || cond.op == OP_LT) {
                        if(ix_compare(cur_max,right_val_ptr,index_col.type,len) > 0){
                            cur_max = right_val_ptr;
                            set_upper = true;
                        }
                    }
                }
            }
            memcpy(lower + current_index, cur_min, len);
            memcpy(upper + current_index, cur_max, len);
            current_index += len;
            delete[] d1;
            delete[] d2;
        }
        left_ = set_lower ? sm_manager_->ihs_.at(index_name_)->lower_bound(lower) : sm_manager_->ihs_.at(index_name_)->leaf_begin();
        right_ = set_upper ? sm_manager_->ihs_.at(index_name_)->upper_bound(upper) : sm_manager_->ihs_.at(index_name_)->leaf_end();

        delete[] lower;
        delete[] upper;
    }

//
//    void get_bound(){
//
//        auto iter = conds_.begin();
//        int cur_offset = 0;
//        //bool is_left_edge_fetched = true;//判断左边界是否能取到等号
//        //bool is_right_edge_fetched = true;//判断右边界是否能取到等号
//        bool is_set_lower = false;
//        bool is_set_upper = false;//可优化？
//        char *lower_key = new char[index_meta_.col_tot_len];
//        char *upper_key = new char[index_meta_.col_tot_len];
//        std::set<std::string> eq_set;
//        init_lower_upper(lower_key,upper_key);
//        for(auto &index_col : index_meta_.cols){
//            //当cond的字段于index_col不匹配时，index_col右移
//            Value prev_g,prev_l;
//            //init_current_min_max(prev_g,prev_l,index_col);//用于id> 3, id > 4的重复比较判断
//            while(iter != conds_.end()){
//                //当与索引字段不匹配；不为右值
//                if(iter->lhs_col.col_name != index_col.name || iter->lhs_col.tab_name != tab_name_){
//                    break;
//                }
//                if(!iter->is_rhs_val){
//                    ++iter;
//                    continue;
//                }
//                switch (iter->op) {
//                    case OP_EQ:
//                        memcpy(lower_key + cur_offset,iter->rhs_val.raw->data,index_col.len);
//                        memcpy(upper_key + cur_offset,iter->rhs_val.raw->data,index_col.len);
//                        is_set_lower = true;
//                        is_set_upper = true;
//                        eq_set.insert(index_col.name);
//                        break;
//                    case OP_GT:
//                        is_left_edge_fetched = false;
//                    case OP_GE:
//                        if(iter->rhs_val > prev_g){
//                            memcpy(lower_key + cur_offset,iter->rhs_val.raw->data,index_col.len);
//                            prev_g = iter->rhs_val;
//                        }
//                        is_set_lower = true;
//                        is_equal = false;
//                        //g_conds_.emplace_back(*iter);
//                        break;
//                    case OP_LT:
//                        is_right_edge_fetched = false;
//                    case OP_LE:
//                        if(iter->rhs_val < prev_l){
//                            memcpy(upper_key + cur_offset,iter->rhs_val.raw->data,index_col.len);
//                            prev_l = iter->rhs_val;
//                        }
//                        is_set_upper = true;
//                        is_equal= false;
//                        //l_conds_.emplace_back(*iter);
//                        break;
//                    default:
//                        assert(0);
//                }
//                ++iter;
//            }
//            cur_offset += index_col.len;
//        }
//
//        is_equal = (eq_set.size() == index_meta_.cols.size());
//
//        //remake_conds();
//
//        //后续针对间隙锁，还需要判断是否所有field都有等号
//        if(is_equal){
//            left_ = ih_->lower_bound(lower_key);
//            right_ = ih_->upper_bound(upper_key);
//            return;
//        }
//        //针对lower部分
//        if(!is_set_lower){
//            left_ = ih_->leaf_begin();
//        }else{
//            left_ = ih_->lower_bound(lower_key);
//        }
//
//        if(!is_set_upper){
//            right_ = ih_->leaf_end();
//        }else{
//            right_ = ih_->upper_bound(upper_key);
//        }
//        delete[] lower_key;
//        delete[] upper_key;
//    }


    void init_lower_upper(char *lower,char *upper){
        int cur_offset = 0;

        //初始化lower和upper
        for(auto &col : cols_){//error
            char *min_type = get_type_min(col);
            char *max_type = get_type_max(col);
            memcpy(lower + cur_offset,min_type,col.len);
            memcpy(upper + cur_offset,max_type,col.len);
            cur_offset += col.len;
            delete[] min_type;
            delete[] max_type;
        }
    }


    //获得过滤后的索引的边界
    //condition中可能包含某一索引列，根据那一列进行过滤，得到b+树的最左和最右的slot情况
    //对于多索引，以第一个非等号符号为OP，对于一个索引的多个<, <=,获取其相关最小值的符号,若值相等，则取<
    //sql情况，表内id: 1 ,2 3, 4, 5,
    //case1:select id = 2, 不用上锁
    //case2:gap_lock在边界的情况:
    //case2.1:select id > 2, id < 4; left应该获取2的后一位，即第一位大于2的值,upper_bound;right应该获取第一个大于等于4的值,lower_bound
    //case2.2:select id >= 2 ,id < 4; left应该获取2,即第一位大于等于2的值,lower_bound;right，同上
    //case2.3:select id >=2 ,id <= 4;left同上，right，第一位大于4的值，5，upper_bound;
    //case2.4:一边没有数字 (如id > 2)，则为end或begin

    void getBound() {
        char *lower = new char[index_meta_.col_tot_len];//还未释放
        char *upper = new char[index_meta_.col_tot_len];

        //bool set_lower = false, set_upper = false;
        // 遍历索引中所有的列，将lower和upper填满
        int current_index = 0;
        CompOp min_cmp = OP_NONE;
        CompOp max_cmp = OP_NONE;//用于记录id<=2还是id<2的符号大小情况，在多维索引中，以第一个列为基准。
        //1:获取该索引列的值
        //2.判断该索引的符号，例如索引id,name,若cond:,name > 'll'，id = 2,则符号为大于号

        for (const auto &index_col : index_meta_.cols) {
            auto index_col_name = index_col.name;
            CompOp index_min_cmp = CompOp::OP_NONE;
            CompOp index_max_cmp = CompOp::OP_NONE;
            Value current_min, current_max; // 初始化最大值和最小值
            // 初始化
//            init_current_min_max(current_min,current_max,index_col);
            auto cond_iter = conds_.begin();
            while (cond_iter != conds_.end()) {
                if(cond_iter->lhs_col.col_name != index_col_name || cond_iter->lhs_col.tab_name != tab_name_ || !cond_iter->is_rhs_val){
                    ++cond_iter;
                    continue;
                }
                //获得索引的值的边界的大小
                auto &rhs_val = cond_iter->rhs_val;
                switch(cond_iter->op){
                    case OP_EQ:
                        if (current_min < rhs_val) {
                            current_min = rhs_val;
                        }
                        if (current_max > rhs_val) {
                            current_max = rhs_val;
                        }
                        index_min_cmp = (index_min_cmp == OP_NONE)?OP_EQ:index_min_cmp;
                        index_max_cmp = (index_max_cmp == OP_NONE)?OP_EQ:index_max_cmp;
                        break;
                    case OP_GE:
                        if (current_min < rhs_val) {
                            current_min = rhs_val;
                            index_min_cmp = OP_GE;
                        }
                        break;
                    case OP_GT:
                        if (!(current_min > rhs_val)) {
                            current_min = rhs_val;
                            index_min_cmp = OP_GT;
                        }
                        break;
                    case OP_LE:
                        if (current_max > rhs_val) {
                            current_max = rhs_val;
                            index_max_cmp = OP_LE;
                        }
                        break;
                    case OP_LT:
                        if (!(current_max < rhs_val)) {
                            current_max = rhs_val;
                            index_max_cmp = OP_LT;
                        }
                        break;
                    default:
                        assert(0);
                }
//                conds_.erase(cond_iter);//保证最后条件里不包含index字段的值比较
//                min_cmp = (min_cmp == CompOp::OP_EQ || min_cmp == CompOp::OP_NONE)?index_min_cmp:min_cmp;
//                max_cmp = (max_cmp == CompOp::OP_EQ || max_cmp == CompOp::OP_NONE)?index_max_cmp:max_cmp;

            }
            auto tmp_len = index_col.len;
            memcpy(lower + current_index, current_min.raw->data, tmp_len);
            memcpy(upper + current_index, current_max.raw->data, tmp_len);
            current_index += tmp_len;
        }

        //cal_pos(min_cmp,max_cmp,lower,upper);
        is_equal = (min_cmp == OP_EQ && max_cmp == OP_EQ) || (min_cmp == OP_NONE && max_cmp == OP_EQ) || (min_cmp == OP_EQ && max_cmp == OP_NONE);
        //等值查询不用间隙锁
        if(is_equal)return;//无需间隙锁
        //left gap获取
        if(min_cmp == OP_NONE || min_cmp == OP_EQ){
            gap_left_ = ih_->leaf_begin();
        }else{
            //assert(0);//不报错
            switch (min_cmp) {
                case OP_GE:{
                    Rid rid;
                    gap_left_ = ih_->upper_bound(lower,rid);
                    break;
                }
                case OP_GT:{
                    Rid rid;
                    gap_left_ = ih_->upper_bound(lower,rid);
                    break;
                }
                default:
                    assert(0);
            }
        }
        //right gap获取
        if(max_cmp == OP_NONE || max_cmp == OP_EQ){
            gap_right_ = ih_->leaf_end();
        }else{
            switch (max_cmp) {
                case OP_LT:{
                    //Rid rid;
                    gap_right_ = ih_->lower_bound(upper);
                    break;
                }
                case OP_LE:{
                    gap_right_ = ih_->lower_bound(upper);
                    break;
                }
                default:
                    assert(0);
            }
        }
        delete[] lower;
        delete[] upper;
    }

    bool check_all_conditions(std::unique_ptr<RmRecord> &record) {
        for (const auto &cond : conds_) {
            if (!check_condition(cols_, cond, record)) {
                return false;
            }
        }
        return true;
    }

    char* get_type_max(ColMeta &col) {
        char* return_ptr = new char[col.len];
        if (col.type ==ColType::TYPE_INT) {
            int max_int = std::numeric_limits<int>::max();
            *(int*)return_ptr = max_int;
        }
        else if (col.type ==ColType::TYPE_FLOAT) {
            float max_float = std::numeric_limits<float>::max();
            *(float*)return_ptr = max_float;
        }else if (col.type ==ColType::TYPE_STRING ) {
            auto max_str = std::string(col.len, 127);
            memset(return_ptr, 0xffff, col.len);
        }
        return return_ptr;
    }

    char* get_type_min(ColMeta &col) {
        char* return_ptr = new char[col.len];
        if (col.type ==ColType::TYPE_INT) {
            int min_int = std::numeric_limits<int>::min();
            *(int*)return_ptr = min_int;
        }
        else if (col.type ==ColType::TYPE_FLOAT ) {
            float float_min = std::numeric_limits<float>::lowest();
            *(float*)return_ptr = float_min;
        }else if (col.type ==ColType::TYPE_STRING) {
            memset(return_ptr, 0x00, col.len);
        }else{
            assert(0);
        }
        return return_ptr;
    }


    bool check_condition(std::vector<ColMeta> &cols, const Condition &cond, std::unique_ptr<RmRecord> &record){
        auto left_col_iter = get_col(cols,cond.lhs_col);

        char *lhs = record->data + left_col_iter->offset;
        char *rhs;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            rhs_type = cond.rhs_val.type;
            rhs = cond.rhs_val.raw->data;
        } else {
            auto rhs_col = get_col(cols,cond.rhs_col);
            rhs_type = rhs_col->type;
            rhs = record->data + rhs_col->offset;
        }

        int cmp = ix_compare(lhs, rhs, rhs_type, left_col_iter->len);
        if (cond.op == OP_EQ) {
            return cmp == 0;
        } else if (cond.op == OP_NE) {
            return cmp != 0;
        } else if (cond.op == OP_LT) {
            return cmp < 0;
        } else if (cond.op == OP_GT) {
            return cmp > 0;
        } else if (cond.op == OP_LE) {
            return cmp <= 0;
        } else if (cond.op == OP_GE) {
            return cmp >= 0;
        } else {
            throw InternalError("Unexpected op type");
        }
    }
};