/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
private:

    //重载abstractExecutor的函数
    //建立BufferSize
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表)）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    size_t left_len_;                          //外表的记录长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段
    std::unique_ptr<RmRecord> merged_rec; // 合并后的rec
    std::vector<Condition> fed_conds_;          // join条件
    std::map<TabCol, Value> prev_feed_dict_;
    char* buffer_join;
    bool is_end_; //State
    int buffer_join_ptr;
    int buffer_stored_cnt;
    std::unique_ptr<RmRecord> right_rec_;

public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right,
            std::vector<Condition> conds){
        //Problem 左右表和内外表的映射关系
        left_ = std::move(left);
        right_ = std::move(right);
        left_len_ = left_->tupleLen();
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        // isend = false;
        fed_conds_ = std::move(conds);
        buffer_join = new char[JOINBLOCK_SIZE];
        is_end_ = false;
        buffer_join_ptr = 0;
    }

    //合并rec
    std::unique_ptr<RmRecord> merge_record(std::unique_ptr<RmRecord>& left_record, std::unique_ptr<RmRecord>& right_record){
        std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>(this->len_);
        //赋值
        memcpy(ret->data, left_record->data, left_record->size);
        if((int)this->len_ > left_record->size){
            memcpy(ret->data + left_record->size, right_record->data, right_record->size);
        }
        return ret;
    }

    ~NestedLoopJoinExecutor(){
        //释放空间
        delete[] buffer_join;
    }

    bool is_end() const {
        return this->is_end_;
    }

    /**
     * 预取left_的rec, 直到buff满或者left全部取完
     * return: prefetch是否成功.left_是否还有数据
    */
    bool preFetchLeft(){
        //将外表的数据存入BUFFER中
        const int max_stored_recnum = JOINBLOCK_SIZE / left_len_;
        int now_strored_recnum = 0;
        buffer_join_ptr = 0;
        if(left_->is_end()){
            this->buffer_stored_cnt = 0;
            return false;
        }
        while(true){
            //取一条record进来
            if(left_->is_end() || now_strored_recnum>= max_stored_recnum){
                break;
            }
            std::unique_ptr<RmRecord> rec = left_->Next();
            //写入buffer_join
            memcpy(buffer_join + now_strored_recnum * left_len_, rec->data,left_len_);
            now_strored_recnum++;
            left_->nextTuple();
        }
        this->buffer_stored_cnt = now_strored_recnum;
        return true;
    }

    void beginTuple() override {
        //先将外表的数据存入BUFFER中
        left_->beginTuple();
        if(left_->is_end()){
            this->is_end_ = true;
            return;
        }
        preFetchLeft();

        //一直取right_的record直到满足样例或者到达末尾
        right_->beginTuple();
        if(right_->is_end()){
            //一开始right_到end,说明right_为空
            this->is_end_ = true;
            return;
        }
        right_rec_ = right_->Next();
        find_next_valid_rec();
    }

    void find_next_valid_rec(){
        //去匹配right
        assert(!right_->is_end());
//        if(right_->is_end()){
//            this->is_end_ = true;
//            return;
//        }
        //right_rec_ = right_->Next();
        while(true){
            //判断终止条件
            if(this->is_end()){
                break;
            }
            std::unique_ptr<RmRecord> left_buff_rec_ = std::make_unique<RmRecord>(left_len_ , buffer_join + buffer_join_ptr * left_len_);
            buffer_join_ptr++;
            if(check_all_conditions(left_buff_rec_,right_rec_)){
                this->merged_rec = std::move(merge_record(left_buff_rec_, right_rec_));
                return;
            }
            if(buffer_join_ptr < this->buffer_stored_cnt)continue;
            //一个左缓存遍历完
            preFetchLeft();
            if(left_->is_end() && this->buffer_stored_cnt == 0){
                //左侧遍历完
                left_->beginTuple();
                preFetchLeft();
                right_->nextTuple();
                if(right_->is_end()){
                    this->is_end_ = true;
                    return;
                }
                right_rec_ = right_->Next();
            }
        }
    }
    void nextTuple() override {
        //left为end且left缓存遍历完时，right右移
        //否则，right不移动，left继续查找
        if(left_->is_end() && (buffer_join_ptr >= this->buffer_stored_cnt)){
            left_->beginTuple();
            preFetchLeft();
            right_->nextTuple();
            if(right_->is_end()){
                this->is_end_ = true;
                return;
            }
            right_rec_ = right_->Next();
        }
        find_next_valid_rec();
    }

    std::unique_ptr<RmRecord> Next() override {
        return std::make_unique<RmRecord>(this->merged_rec->size, this->merged_rec->data);
    }

    Rid &rid() override { return _abstract_rid; }

    const std::vector<ColMeta> &cols() const {
        return this->cols_;
    };

    size_t tupleLen() const {
        return this->len_;
    }

    std::string getType() { return "BlockNestedJoinExecutor"; }

    ColMeta get_col_offset(const TabCol &target) override{
        for(auto iter = this->cols_.begin(); iter!=this->cols_.end(); iter++){
            if(iter->tab_name == target.tab_name && iter->name == target.col_name){
                return *iter;
            }
        }
        return ColMeta();
    };

    bool check_all_conditions(std::unique_ptr<RmRecord> &left_record,std::unique_ptr<RmRecord> &right_record) {

        for (const auto &cond : fed_conds_) {
            auto left_col_iter = left_->get_col(left_->cols(),cond.lhs_col);
            //得到左子树的值
            Value left_value;
            try{
                left_value = get_value(left_record,*left_col_iter);
            }catch (DeBugError &e){
                std::cout<<"failed at left vale"<<std::endl;
                std::cout << "this first col;" <<this->cols_[0]<< " this second col"<<std::endl;
                assert(0);
            }
            assert(!cond.is_rhs_val);
            //assert(left_value.type == cond.rhs_col)
            // 找条件中的右边一列
            Value right_value;
            if(cond.is_rhs_sub_query) {
                auto col = right_->cols()[0];
                col.offset = 0;
                right_value = get_value(right_record,col);
            }else if (cond.is_rhs_val) {
                right_value = cond.rhs_val;
            } else {
                auto right_col_iter = right_->get_col(right_->cols(),cond.rhs_col);
                try{
                    right_value = get_value(right_record,*right_col_iter);
                }catch (DeBugError &e){
                    std::cout<<"failed at right vale"<<std::endl;
                    std::cout << "this first col;" <<this->cols_[0]<< " this second col"<<std::endl;
                    assert(0);
                }

            }
            //获取到左右两个值，可以互相比较
            //当不匹配时马上false
            if(!compare_value(left_value,right_value,cond.op))return false;
        }
        return true;
    }

    bool compare_value(const Value& left_value,const Value& right_value,CompOp op){
        switch (op){
            case OP_EQ:
            case OP_IN:
                return left_value == right_value;
            case OP_NE:
                return left_value != right_value;
            case OP_LT:
                return left_value < right_value;
            case OP_GT:
                return left_value > right_value;
            case OP_LE:
                return !(left_value > right_value);
            case OP_GE:
                return !(left_value < right_value);
            default:
                break;
        }
        return false;
    }

    bool eval_cond(const std::vector<ColMeta> &rec_cols, const Condition &cond, std::unique_ptr<RmRecord>& rec) {
        auto lhs_col = get_col(rec_cols, cond.lhs_col);
        char *lhs = rec->data + lhs_col->offset;
        char *rhs;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            rhs_type = cond.rhs_val.type;
            rhs = cond.rhs_val.raw->data;
        } else {
            auto rhs_col = get_col(rec_cols, cond.rhs_col);
            rhs_type = rhs_col->type;
            rhs = rec->data + rhs_col->offset;
        }
        assert(rhs_type == lhs_col->type);
        int cmp = ix_compare(lhs, rhs, rhs_type, lhs_col->len);
        if (cond.op == OP_EQ) {
            return cmp == 0;
        } else if (cond.op == OP_NE) {
            return cmp != 0;
        } else if (cond.op == OP_LT) {
            return cmp < 0;
        } else if (cond.op == OP_GT) {
            return cmp > 0;
        } else if (cond.op == OP_LE) {
            return cmp <= 0;
        } else if (cond.op == OP_GE) {
            return cmp >= 0;
        } else {
            throw InternalError("Unexpected op type");
        }
    }
    bool eval_conds(const std::vector<ColMeta> &rec_cols, const std::vector<Condition> &conds, std::unique_ptr<RmRecord>& rec) {
        for(auto &cond: conds) {
            if(!eval_cond(rec_cols, cond, rec)) {
                return false;
            }
        }
        return true;
    }
};