/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    RmFileHandle *fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

    Rid rid_;
    std::unique_ptr<RecScan> scan_;     // table_iterator
    std::unique_ptr<RmRecord> record_buffer;
    bool is_first_sub_value;
    SmManager *sm_manager_;

public:
    SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;

        context_ = context;
        fed_conds_ = conds_;
        //context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
//        if(conds_.empty()){
//            context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
//        }else{
//            context_->lock_mgr_->lock_IS_on_table(context_->txn_, fh_->GetFd());
//        }
        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
    }

    bool check_all_conditions(std::unique_ptr<RmRecord> &record) {
        for (auto &cond : fed_conds_) {
            if (!check_condition(cols_, cond, record)) {
                return false;
            }
        }
        return true;
    }

    bool check_condition(std::vector<ColMeta> &cols, Condition &cond, std::unique_ptr<RmRecord> &record) {
        // 没有tabs所以要传一个cols
        // 要把cond中的列的位置找出来
        auto left_col_name = cond.lhs_col.col_name;
        ColMeta left_col, right_col;
        for (const auto &col : cols) {
            if (col.name == left_col_name && col.tab_name == tab_name_) {
                left_col = col;
                break;
            }
        }
        auto left_value = get_value(record, left_col);
        // 找条件中的右边一列

        Value right_value;
        if (cond.is_rhs_val) {
            if(!cond.is_rhs_sub_query){
                right_value = cond.rhs_val;
            }else{
                get_sub_value(cond);
                right_value = cond.rhs_val;
            }
        }else{
            for (const auto &col : cols) {
                if (col.name == cond.rhs_col.col_name && col.tab_name == cond.rhs_col.tab_name) {
                    right_col = col;
                    break;
                }
            }
            try{
                right_value = get_value(record, right_col);
            }catch (DeBugError &e){
                std::cout << "seqscasn error"<<std::endl;
                assert(0);
            }

        }

        return compare_value(left_value, right_value, cond.op);
    }

    void get_sub_value(Condition &cond){
        Value sub_value;
        auto sub_exec = std::move(cond.sub_query_info->sub_query_exec);
        if(sub_exec == nullptr)return;//保证只执行一次后，存储子查询值
        for (sub_exec->beginTuple(); !sub_exec->is_end(); sub_exec->nextTuple()) {
            auto rec = sub_exec->Next();
            auto value = get_value(rec,sub_exec->cols()[0]);//需保证只有一列
            sub_value.add_to_set(value);
        }
        cond.rhs_val = sub_value;
    }

    bool compare_value(const Value& left_value,const Value& right_value,CompOp op){
        switch (op){
            case OP_EQ:
                return left_value == right_value;
            case OP_NE:
                return left_value != right_value;
            case OP_LT:
                return left_value < right_value;
            case OP_GT:
                return left_value > right_value;
            case OP_LE:
                return !(left_value > right_value);
            case OP_GE:
                return !(left_value < right_value);
            case OP_IN:
                return right_value.is_in_set(left_value);
            default:
                assert(0);
        }
    }

    /**
     * 这里begin和next一样，区别？
     */
    void beginTuple() override {
        scan_ = std::make_unique<RmScan>(fh_);
        while(!scan_->is_end()) {
            // 遍历
            rid_ = scan_->rid();
            if(conds_.empty())break;
            record_buffer = fh_->get_record(rid_, nullptr);
            bool check = check_all_conditions(record_buffer);
            if (check) {
                rid_ = scan_->rid();
                return;
            }
            scan_->next();
        }
    }

    void nextTuple() override {
        //assert(!is_end());
        scan_->next();
        rid_ = scan_->rid();
        if(conds_.empty())return;
        while (!scan_->is_end()) {
            record_buffer = fh_->get_record(scan_->rid(), nullptr);
            bool check = check_all_conditions(record_buffer);
            if (check) {
                rid_ = scan_->rid();
                return;
            }
            scan_->next();
        }
    }

    size_t tupleLen() const override { return len_; }

    std::string getType() override { return "SeqScanExecutor"; }

    const std::vector<ColMeta> &cols() const override { return cols_; }

    /**
     * 根据next tuple，获得rid_
     * @return
     */
    std::unique_ptr<RmRecord> Next() override {
        assert(!is_end());
        return fh_->get_record(rid_, nullptr);
    }

    bool is_end() const override { return scan_->is_end(); }

    Rid &rid() override { return rid_; }

    ColMeta get_col_offset(const TabCol &target) override {
        ColMeta result;
        for (const auto &col : cols_) {
            if (col.tab_name == target.tab_name && col.name == target.col_name) {
                result = col;
                break;
            }
        }
        return result;
    }
};

