//
// Created by ll on 8/7/24.
//

#ifndef RMDB_EXECUTOR_SUB_INDEX_SCAN_H
#define RMDB_EXECUTOR_SUB_INDEX_SCAN_H
/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SubIndexScanExecutor : public AbstractExecutor {
private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    std::vector<Condition> conds_;              // 扫描条件
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    //std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<IxScan> scan_;


    SmManager *sm_manager_;

    char *lower_key;                            //根据conds，获得的键，索引扫描时的下界，不包括index的第一个字段
    char *upper_key;                            //根据conds获得的键，索引扫描时的上界，不包括index的第一个字段
    char *prob_next_key;                        //用于获得下一个index的第一个字段值，其他字段值全部赋值为最大值
    bool end;
//    CompOp min_cmp;
//    CompOp max_cmp;
    Iid left_;
    Iid right_;
    int cur_id;
    int final_first_field;
    std::string index_name_;
    std::unique_ptr<RmRecord> record_buffer;
    IxIndexHandle * ih_;
public:
    SubIndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string> index_col_names,
                      Context *context) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        conds_ = std::move(conds);  //sorted conditions,index related condition is first,后续只剩下非index条件
        index_col_names_ = index_col_names;
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        end = false;
        index_name_ = sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta_.cols);
        ih_ = sm_manager_->ihs_.at(index_name_).get();
        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());//暂时用大锁，不用间隙锁
        lower_key = nullptr;
        upper_key = nullptr;
        prob_next_key = nullptr;
        //assert(0);
        //获得lower_key和upper_key,不对index的第一个字段赋值
        getBound();
    }

    ~SubIndexScanExecutor() override {
        delete[] lower_key;
        delete[] upper_key;
        delete[] prob_next_key;
//        if(lower_key != nullptr){
//            delete[] lower_key;
//        }
//        if(upper_key != nullptr){
//            delete[] upper_key;
//        }
//        if(prob_next_key != nullptr){
//            delete[] prob_next_key;
//        }
    }

    void beginTuple() override {
        while(cur_id <= final_first_field){
            //第一个索引遍历
            if(check_in_one_id_range()){
                //在当前范围内找到匹配值
                break;
            }
            bool is_get_next = get_next_curr_ind();
            if(!is_get_next){
                end = true;
                break;
            }
            *(int*) lower_key = cur_id;
            *(int*) upper_key = cur_id;
            cal_pos();
            //重新生成scan
            scan_->remake(left_,right_);
        }
    }

    void nextTuple() override {
        scan_->next();
        while(cur_id <= final_first_field){
            //第一个索引遍历
            if(check_in_one_id_range()){
                //在当前范围内找到匹配值
                break;
            }
            bool is_get_next = get_next_curr_ind();
            if(!is_get_next){
                end = true;
                break;
            }
            *(int*) lower_key = cur_id;
            *(int*) upper_key = cur_id;//可能冗余计算
            cal_pos();
            //重新生成scan
            scan_->remake(left_,right_);
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        return std::move(record_buffer);
    }

    size_t tupleLen() const override { return len_; }

    std::string getType() override { return "IndexScanExecutor"; }

    const std::vector<ColMeta> &cols() const override { return cols_; }

    bool is_end() const override {
        bool is_end = scan_->is_end();
        return end || is_end;
    }

    ColMeta get_col_offset(const TabCol &target) override {
        ColMeta result;
        for (const auto &col : cols_) {
            if (col.tab_name == target.tab_name && col.name == target.col_name) {
                result = col;
                break;
            }
        }
        return result;
    }

    Rid &rid() override { return rid_; }


    void init_prob_key(){
        prob_next_key = new char[index_meta_.col_tot_len];//注意后续delete
        auto iter = index_meta_.cols.begin() + 1;
        int offset = sizeof(int);
        for(;iter != index_meta_.cols.end();++iter){
            char *absolute_max = get_type_max(*iter);
            memcpy(prob_next_key + offset,absolute_max,iter->len);
            offset += iter->len;
            delete[] absolute_max;//优化
        }
    }

    void init_lower_upper_key(){
        lower_key = new char[index_meta_.col_tot_len];//还未释放
        upper_key = new char[index_meta_.col_tot_len];
        int current_index = sizeof(int);
        char *cur_max;
        char *cur_min;
        for(auto index_col_iter = index_meta_.cols.begin() + 1; index_col_iter != index_meta_.cols.end();++index_col_iter){
            auto index_col_name = index_col_iter->name;
            cur_max = get_type_max(*index_col_iter);
            cur_min = get_type_min(*index_col_iter);
            char *d1 = cur_max;
            char *d2 = cur_min;
            //auto cond_iter = conds_.begin();
            for (const auto &cond : conds_) {
                if(!cond.is_rhs_val)continue;
                //note：需保证cond内不出现其他表的cond
                if (cond.lhs_col.col_name == index_col_name) {
                    //假设此时cond内不包含其他表的信息
                    // 找到了条件中的指定列
                    char *right_val_ptr = cond.rhs_val.raw->data;
                    if (cond.op == OP_EQ) {
                        if(ix_compare(cur_min,right_val_ptr,index_col_iter->type,index_col_iter->len) < 0){
                            cur_min = right_val_ptr;
                        }
                        if(ix_compare(cur_max,right_val_ptr,index_col_iter->type,index_col_iter->len) > 0){
                            cur_max = right_val_ptr;
                        }
                    } else if (cond.op == OP_GE || cond.op == OP_GT) {
                        if(ix_compare(cur_min,right_val_ptr,index_col_iter->type,index_col_iter->len) < 0){
                            cur_min = right_val_ptr;
                        }
                    } else if (cond.op == OP_LE || cond.op == OP_LT) {
                        if(ix_compare(cur_max,right_val_ptr,index_col_iter->type,index_col_iter->len) > 0){
                            cur_max = right_val_ptr;
                        }
                    }
                }
            }
            memcpy(lower_key + current_index, cur_min, index_col_iter->len);
            memcpy(upper_key + current_index, cur_max, index_col_iter->len);
            current_index += index_col_iter->len;
            delete[] d1;
            delete[] d2;//TODO 后续修改为unique_ptr管理内存
        }
    }

    void cal_pos(){
        left_ = ih_->lower_bound(lower_key);
        right_ = ih_->upper_bound(upper_key);
    }

    //获得过滤后的索引的边界
    void getBound() {
        init_prob_key();
        init_lower_upper_key();
        cur_id = ih_->first_leaf_first_key();
        final_first_field = ih_->last_leaf_last_key();
        if(cur_id > final_first_field){
            //值不存在
            scan_ = std::make_unique<IxScan>(ih_, Iid{-1,-1}, Iid{-1,-1}, sm_manager_->get_bpm());
            end = true;
            return;
        }
        *(int *)lower_key = cur_id;
        *(int *)upper_key = cur_id;//第一个缺失index字段只能是int类型
        cal_pos();
        scan_ = std::make_unique<IxScan>(ih_, left_, right_, sm_manager_->get_bpm());
    }

    bool get_next_curr_ind(){
        int tmp;
        *(int*) prob_next_key = cur_id;
        Iid next_pos = ih_->upper_bound(prob_next_key);//暂时不支持上锁
        //next_pos case:1,到了end，表示中止，
        //case2:不为end，复制
        if(ih_->find_ind_key_at(next_pos, tmp)){
            cur_id = tmp;
            std::cout<< "tmp: "<<tmp<<std::endl;
            return true;
        }else{
            return false;
        }
    }

    bool check_in_one_id_range(){
        while(!scan_->is_end()) {
            // 遍历
            rid_ = scan_->rid();
            record_buffer = fh_->get_record(rid_, nullptr);//TODO 里面加锁，隐患：left和right的边界没有确定，可能会有多余的锁
            //可能存在冗余检查，因为之前索引部分的条件依据通过过滤得到
            bool check = check_all_conditions(record_buffer);
            if (check) {
                rid_ = scan_->rid();
                return true;
            }
            scan_->next();
        }
        return false;
    }
    bool check_all_conditions(std::unique_ptr<RmRecord> &record) {
        for (const auto &cond :conds_) {
            if (!check_condition(cols_, cond, record)) {
                return false;
            }
        }
        return true;
    }

    char* get_type_max(ColMeta &col) {
        char* return_ptr = new char[col.len];
        if (col.type ==ColType::TYPE_INT) {
            int max_int = std::numeric_limits<int>::max();
            *(int*)return_ptr = max_int;
        }
        else if (col.type ==ColType::TYPE_FLOAT) {
            float max_float = std::numeric_limits<float>::max();
            *(float*)return_ptr = max_float;
        }else if (col.type ==ColType::TYPE_STRING ) {
            auto max_str = std::string(col.len, 127);
            memset(return_ptr, 0xffff, col.len);
        }
        return return_ptr;
    }

    char* get_type_min(ColMeta &col) {
        char* return_ptr = new char[col.len];
        if (col.type ==ColType::TYPE_INT) {
            int min_int = std::numeric_limits<int>::min();
            *(int*)return_ptr = min_int;
        }
        else if (col.type ==ColType::TYPE_FLOAT ) {
            float float_min = std::numeric_limits<float>::lowest();
            *(float*)return_ptr = float_min;
        }else if (col.type ==ColType::TYPE_STRING) {
            memset(return_ptr, 0x00, col.len);
        }else{
            assert(0);
        }
        return return_ptr;
    }


    bool check_condition(std::vector<ColMeta> &cols, const Condition &cond, std::unique_ptr<RmRecord> &record){
        auto left_col_iter = get_col(cols,cond.lhs_col);

        char *lhs = record->data + left_col_iter->offset;
        char *rhs;
        ColType rhs_type;
        if (cond.is_rhs_val) {
            rhs_type = cond.rhs_val.type;
            rhs = cond.rhs_val.raw->data;
        } else {
            auto rhs_col = get_col(cols,cond.rhs_col);
            rhs_type = rhs_col->type;
            rhs = record->data + rhs_col->offset;
        }

        int cmp = ix_compare(lhs, rhs, rhs_type, left_col_iter->len);
        if (cond.op == OP_EQ) {
            return cmp == 0;
        } else if (cond.op == OP_NE) {
            return cmp != 0;
        } else if (cond.op == OP_LT) {
            return cmp < 0;
        } else if (cond.op == OP_GT) {
            return cmp > 0;
        } else if (cond.op == OP_LE) {
            return cmp <= 0;
        } else if (cond.op == OP_GE) {
            return cmp >= 0;
        } else {
            throw InternalError("Unexpected op type");
        }
    }

};
#endif //RMDB_EXECUTOR_SUB_INDEX_SCAN_H
