/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;
    bool isend;
    std::unique_ptr<RmRecord> record;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = std::move(set_clauses);
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;//seq_scan获取的相关行
        context_ = context;
        isend = false;

        if (!context_->lock_mgr_->lock_IX_on_table(context_->txn_, fh_->GetFd())) {
            // 拿不到
            throw TransactionAbortException(context_->txn_->get_transaction_id(),AbortReason::UPGRADE_CONFLICT);
        }
        //context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd());
    }
    std::unique_ptr<RmRecord> Next() override {
        // 遍历所有的记录
        //模拟火山模型，一条条改
        for (auto &rid : rids_) {
            //TODO 聚簇型的优化读取，使用缓冲页
            record = fh_->get_record(rid, context_);//record为旧数据
            char *old_rec = new char[record->size];//备份旧数据
            memcpy(old_rec, record->data, record->size);
            char *buf = new char[record->size];//存放更新后的行，最后由两部分构成：修改后的value和不涉及修改的value
            memcpy(buf,old_rec,record->size);
            //从set_clauses,获取set value中的value，存在buffer中
            for (auto &clause : set_clauses_) {
                update_record(buf, clause);
            }

            // 检查索引重复
            for (size_t i = 0; i < tab_.indexes.size(); ++i) {
                auto &index = tab_.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                char* key = new char[index.col_tot_len];
                char* prev = new char[index.col_tot_len];
                int offset = 0;
                std::vector<ColType> types;
                std::vector<int> lens;
                for(size_t j = 0; j < (size_t)index.col_num; ++j) {
                    memcpy(key + offset, buf + index.cols[j].offset, index.cols[j].len);
                    memcpy(prev + offset, record->data + index.cols[j].offset, index.cols[j].len);
                    offset += index.cols[j].len;
                    types.push_back(index.cols[j].type);
                    lens.push_back(index.cols[j].len);
                }
                if (!ix_compare(key, prev, types, lens)) {
                    continue;
                }
                std::vector<Rid> result;
                ih->lock_root();
                if (ih->get_value(key, &result, context_->txn_)) {
                    ih->unlock_root();
                    throw DuplicateKeyError();
                }
                ih->unlock_root();
                delete[] key;
                delete[] prev;
            }

            //写操作写入txn
            auto write_record = new WriteRecord(WType::UPDATE_TUPLE, tab_name_, rid, *record);//存储旧值
            std::unique_ptr<RmRecord> new_rec(new RmRecord(record->size,buf));
            context_->txn_->append_write_record(write_record);
            if(enable_logging){
                auto update_log(new UpdateLogRecord(context_->txn_->get_transaction_id(),context_->txn_->get_prev_lsn(),
                                                    *record,*new_rec,rid,tab_name_));
                context_->log_mgr_->add_log_to_buffer(update_log);
                //记录最新修改的page的lsn
                Page* page = sm_manager_->get_bpm()->fetch_page({fh_->GetFd(), rid.page_no});
                page->set_page_lsn(update_log->lsn_);
                sm_manager_->get_bpm()->unpin_page({fh_->GetFd(), rid.page_no}, true);
            }


            // Insert into index
            for(size_t i = 0; i < tab_.indexes.size(); ++i) {
                auto& index = tab_.indexes[i];
                auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();

                char* old_key = new char[index.col_tot_len];
                char* new_key = new char[index.col_tot_len];
                int offset = 0;
                for(size_t j = 0; j < (size_t)index.col_num; ++j) {
                    memcpy(old_key + offset, write_record->GetRecord().data + index.cols[j].offset, index.cols[j].len);
                    memcpy(new_key + offset, buf + index.cols[j].offset,index.cols[j].len);
                    offset += index.cols[j].len;
                }
                //ih->delete_entry(old_key, context_->txn_);
                //间隙锁
//                if(context_->txn_->get_isolation_level() == IsolationLevel::SERIALIZABLE){
//                    Rid gap_rid = {-1,-1};
//                    ih->upper_bound(new_key,gap_rid);
//                    context_->lock_mgr_->try_lock_gap_on_index(context_->txn_,ih->get_fd(),gap_rid);
//                }

                //old_keys.emplace_back(old_key);
                //new_keys.emplace_back(new_key);//可能存在的问题：key太多，char*大小不确定，vector内存爆炸
                //TODO 性能开销权衡 IF old_key = new_key

                //ih->lock_root();
                ih->delete_entry(old_key, context_->txn_);//TODO 这里是大锁，可以后续优化
                ih->insert_entry(new_key, rid, context_->txn_);
                //ih->unlock_root();
                delete[] new_key;
                delete[] old_key;
            }
            fh_->update_record(rid, buf, nullptr);//TODO get_record时已经加锁
            //上一步已经保证新的insert值获取到相应间隙锁，统一加锁
            //fh_->update_record(rid, buf, nullptr);//TODO:update的间隙锁可能存在比较大的隐患
            //获取到修改值的间隙锁，
//            for(size_t i = 0; i < tab_.indexes.size(); ++i) {
//                auto &index = tab_.indexes[i];
//                auto ih = sm_manager_->ihs_.at(
//                        sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
////                if(context_->txn_->get_isolation_level() == IsolationLevel::SERIALIZABLE){
////                    //调整间隙
////                    Rid gap_rid;
////                    ih->upper_bound(old_keys[i],gap_rid);
////                    context_->lock_mgr_->try_erase_gap_on_index(context_->txn_,ih->get_fd(),gap_rid);//删除同一事务的gap
////                }
////if(old_key == new_key)continue;
//                ih->delete_entry(old_keys[i], context_->txn_);
//                ih->insert_entry(new_keys[i], rid, context_->txn_);
//            }

            delete[] buf;
            delete[] old_rec;
        }
        isend = true;
        return nullptr;
    }

    size_t tupleLen() const override { return fh_->get_record(rids_[0], context_)->size; }

    std::string getType() override { return "UpdateExecutor"; }

    bool is_end() const override { return isend; }

    Rid &rid() override { return _abstract_rid; }

    ColMeta get_col_offset(const TabCol &target) override {
        ColMeta result;
        for (const auto &clause : set_clauses_) {
            if (clause.lhs.col_name == target.col_name && clause.lhs.tab_name == target.tab_name) {
                result = *(tab_.get_col(clause.lhs.col_name));
                break;
            }
        }
        return result;
    }

    void update_record(char *buf, SetClause &clause) {
        auto col = tab_.get_col(clause.lhs.col_name);
        int len = col->len, offset = col->offset;
        Value value = clause.rhs;
        std::string val2;
        switch (col->type)
        {
        case TYPE_INT:
            int val;
            if (value.type == TYPE_FLOAT) {
                val = (int)value.float_val;
            } else {
                val = value.int_val;
            }

            if (clause.op == NO) {
                *(int *)(buf + offset) = val;
            } else if (clause.op == PLUS) {
                *(int *)(buf + offset) += val;
            } else if (clause.op == MINUS) {
                *(int *)(buf + offset) -= val;
            } else if (clause.op == MULTI) {
                *(int *)(buf + offset) *= val;
            } else if (clause.op == DIVIDE) {
                *(int *)(buf + offset) /= val;
            }

            break;
        case TYPE_FLOAT:
            float val1;
            if (value.type == TYPE_INT) {
                val1 = (float)value.int_val;
            } else {
                val1 = value.float_val;
            }

            if (clause.op == NO) {
                *(float *)(buf + offset) = val1;
            } else {
                float result = *(float *)(buf + offset);
                if (clause.op == PLUS) {
                    result += val1;
                } else if (clause.op == MINUS) {
                    result -= val1;
                } else if (clause.op == MULTI) {
                    result *= val1;
                } else if (clause.op == DIVIDE) {
                    result /= val1;
                }
                *(float *)(buf + offset) = result;
            }

            break;
        case TYPE_STRING:
            memcpy(buf + offset, value.str_val.c_str(), len);
            break;
        default:
            break;
        }
    }

};