/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */
#pragma once

#include <vector>
#include <string>
#include <memory>
#include <iostream>
enum JoinType {
    INNER_JOIN, LEFT_JOIN, RIGHT_JOIN, FULL_JOIN
};
namespace ast {

enum SvType {
    SV_TYPE_INT, SV_TYPE_FLOAT, SV_TYPE_STRING, SV_TYPE_BOOL,SV_TYPE_DATETIME
};

enum SvAgreType {
    TYPE_MIN = 0, TYPE_MAX, TYPE_SUM,TYPE_COUNT,TYPE_COUNT_ALL
};

enum SvCompOp {
    SV_OP_EQ, SV_OP_NE, SV_OP_LT, SV_OP_GT, SV_OP_LE, SV_OP_GE,SV_OP_IN
};

enum OrderByDir {
    OrderBy_DEFAULT,
    OrderBy_ASC,
    OrderBy_DESC
};

enum SetKnobType {
    EnableNestLoop, EnableSortMerge
};

enum SvUpdateOp {
    SV_NULL = 0, SV_ADD, SV_MINUS, SV_MULTI, SV_DIVIDE
};

// Base class for tree nodes
struct TreeNode {
    virtual ~TreeNode() = default;  // enable polymorphism
};

struct Help : public TreeNode {
};

struct ShowTables : public TreeNode {
};

struct TxnBegin : public TreeNode {
};

struct TxnCommit : public TreeNode {
};

struct TxnAbort : public TreeNode {
};

struct TxnRollback : public TreeNode {
};

struct StaticCheckPoint : public TreeNode{
};

struct TypeLen : public TreeNode {
    SvType type;
    int len;

    TypeLen(SvType type_, int len_) : type(type_), len(len_) {}
};

struct Field : public TreeNode {
};

struct ColDef : public Field {
    std::string col_name;
    std::shared_ptr<TypeLen> type_len;

    ColDef(std::string col_name_, std::shared_ptr<TypeLen> type_len_) :
            col_name(std::move(col_name_)), type_len(std::move(type_len_)) {}
};

struct CreateTable : public TreeNode {
    std::string tab_name;
    std::vector<std::shared_ptr<Field>> fields;

    CreateTable(std::string tab_name_, std::vector<std::shared_ptr<Field>> fields_) :
            tab_name(std::move(tab_name_)), fields(std::move(fields_)) {}
};

struct DropTable : public TreeNode {
    std::string tab_name;

    DropTable(std::string tab_name_) : tab_name(std::move(tab_name_)) {}
};

struct DescTable : public TreeNode {
    std::string tab_name;

    DescTable(std::string tab_name_) : tab_name(std::move(tab_name_)) {}
};

struct CreateIndex : public TreeNode {
    std::string tab_name;
    std::vector<std::string> col_names;

    CreateIndex(std::string tab_name_, std::vector<std::string> col_names_) :
            tab_name(std::move(tab_name_)), col_names(std::move(col_names_)) {}
};

struct DropIndex : public TreeNode {
    std::string tab_name;
    std::vector<std::string> col_names;

    DropIndex(std::string tab_name_, std::vector<std::string> col_names_) :
            tab_name(std::move(tab_name_)), col_names(std::move(col_names_)) {}
};

struct ShowIndex : public TreeNode {
    std::string tab_name;

    ShowIndex(std::string tab_name_) :
            tab_name(std::move(tab_name_)) {}
};


struct Expr : public TreeNode {
};

struct Value : public Expr {
};

struct IntLit : public Value {
    int val;

    IntLit(int val_) : val(val_) {}
};

struct FloatLit : public Value {
    float val;

    FloatLit(float val_) : val(val_) {}
};

struct StringLit : public Value {
    std::string val;

    StringLit(std::string val_) : val(std::move(val_)) {}
};

struct BoolLit : public Value {
    bool val;

    BoolLit(bool val_) : val(val_) {}
};

struct AgreInfo{
    SvAgreType agre_type;
    std::string agre_name;
    AgreInfo(SvAgreType agre_type_, std::string agre_name_) :
    agre_type(agre_type_), agre_name(std::move(agre_name_)) {
    }
};

struct Col : public Expr {
    std::string tab_name;
    std::string col_name;
    bool is_aggregate;
    SvAgreType agre_type;
    std::string alias;

    Col(std::string tab_name_, std::string col_name_) :
            tab_name(std::move(tab_name_)), col_name(std::move(col_name_)) {
        is_aggregate = false;
    }

    Col(SvAgreType agre_type_,std::string tab_name_, std::string col_name_,std::string alias_) :
            tab_name(std::move(tab_name_)),
            col_name(std::move(col_name_)),
            agre_type(agre_type_),
            alias(std::move(alias_)) {
        is_aggregate = true;
    }

};



struct SetClause : public TreeNode {
    std::string col_name;
    std::shared_ptr<Value> val;
    SvUpdateOp op;

    SetClause(std::string col_name_, std::shared_ptr<Value> val_, SvUpdateOp op_ = SV_NULL) :
            col_name(std::move(col_name_)), val(std::move(val_)), op(std::move(op_)) {}
};

struct BinaryExpr : public TreeNode {

    std::shared_ptr<Col> lhs;
    SvCompOp op;
    std::shared_ptr<Expr> rhs;
    bool is_having;

    BinaryExpr(std::shared_ptr<Col> lhs_, SvCompOp op_, std::shared_ptr<Expr> rhs_) :
            lhs(std::move(lhs_)), op(op_), rhs(std::move(rhs_)) {
        is_having = false;
    }
};

struct OrderBy : public TreeNode
{
    std::shared_ptr<Col> cols;
    OrderByDir orderby_dir;
    OrderBy( std::shared_ptr<Col> cols_, OrderByDir orderby_dir_) :
       cols(std::move(cols_)), orderby_dir(std::move(orderby_dir_)) {}
};


struct InsertStmt : public TreeNode {
    std::string tab_name;
    std::vector<std::shared_ptr<Value>> vals;

    InsertStmt(std::string tab_name_, std::vector<std::shared_ptr<Value>> vals_) :
            tab_name(std::move(tab_name_)), vals(std::move(vals_)) {}
};

struct DeleteStmt : public TreeNode {
    std::string tab_name;
    std::vector<std::shared_ptr<BinaryExpr>> conds;

    DeleteStmt(std::string tab_name_, std::vector<std::shared_ptr<BinaryExpr>> conds_) :
            tab_name(std::move(tab_name_)), conds(std::move(conds_)) {}
};

struct UpdateStmt : public TreeNode {
    std::string tab_name;
    std::vector<std::shared_ptr<SetClause>> set_clauses;
    std::vector<std::shared_ptr<BinaryExpr>> conds;

    UpdateStmt(std::string tab_name_,
               std::vector<std::shared_ptr<SetClause>> set_clauses_,
               std::vector<std::shared_ptr<BinaryExpr>> conds_) :
            tab_name(std::move(tab_name_)), set_clauses(std::move(set_clauses_)), conds(std::move(conds_)) {}
};

struct JoinExpr : public TreeNode {
    std::string left;
    std::string right;
    std::vector<std::shared_ptr<BinaryExpr>> conds;
    JoinType type;

    JoinExpr(std::string left_, std::string right_,
               std::vector<std::shared_ptr<BinaryExpr>> conds_, JoinType type_) :
            left(std::move(left_)), right(std::move(right_)), conds(std::move(conds_)), type(type_) {}
};

struct SelectStmt : public TreeNode {
    std::vector<std::shared_ptr<Col>> cols;
    //std::vector<std::shared_ptr<AgreCol>> agre_cols;
    std::vector<std::string> tabs;
    std::vector<std::shared_ptr<BinaryExpr>> conds;
    std::vector<std::shared_ptr<JoinExpr>> jointree;
    std::vector<std::shared_ptr<Col>> group_by_cols; // Group By Columns
    std::vector<std::shared_ptr<BinaryExpr>> having_conds; // Having Conditions

    bool has_sort;
    std::shared_ptr<OrderBy> order;


    SelectStmt(std::vector<std::shared_ptr<Col>> cols_,
               std::vector<std::string> tabs_,
               std::vector<std::shared_ptr<BinaryExpr>> conds_,
               std::shared_ptr<OrderBy> order_) :
            cols(std::move(cols_)), tabs(std::move(tabs_)), conds(std::move(conds_)), 
            order(std::move(order_)) {
                has_sort = (bool)order;
            }

//    SelectStmt(std::vector<std::shared_ptr<Col>> agre_cols_,
//               std::vector<std::string> tabs_,
//               std::vector<std::shared_ptr<BinaryExpr>> conds_,
//               std::shared_ptr<OrderBy> order_) :
//            agre_cols(std::move(agre_cols_)), tabs(std::move(tabs_)), conds(std::move(conds_)),
//            order(std::move(order_)) {
//        //std::cout << "SelectStmt: " << agre_cols_. << std::endl;
//        has_sort = (bool)order;
//    }

    SelectStmt(std::vector<std::shared_ptr<Col>> cols_,
               std::vector<std::string> tabs_,
               std::vector<std::shared_ptr<BinaryExpr>> conds_,
               std::vector<std::shared_ptr<Col>> group_by_cols_,
               std::vector<std::shared_ptr<BinaryExpr>> having_conds_,
               std::shared_ptr<OrderBy> order_):
            cols(std::move(cols_)),
            tabs(std::move(tabs_)),
            conds(std::move(conds_)),
            group_by_cols(std::move(group_by_cols_)),
            having_conds(std::move(having_conds_)),
            order(std::move(order_))
            {
        has_sort = (bool)order;
    }

};

// 新增子查询节点定义
class SubQueryExpr : public Expr {
public:
    std::shared_ptr<SelectStmt> subquery;
    std::vector<std::shared_ptr<Value>> const_values;
    bool is_const;

    SubQueryExpr(std::shared_ptr<SelectStmt> subquery_)
        : subquery(std::move(subquery_)) {
        is_const = false;
    }

    SubQueryExpr( std::vector<std::shared_ptr<Value>> const_){
        is_const = true;
        const_values = std::move(const_);
    }
};

//class InExpr : public Expr {
//public:
//    std::shared_ptr<Col> column;
//    std::shared_ptr<SubQueryExpr> subquery;
//
//    InExpr(std::shared_ptr<Col> col, std::shared_ptr<SubQueryExpr> subq)
//        : column(std::move(col)), subquery(std::move(subq)) {}
//
//    // Implement necessary methods here
//};


// set enable_nestloop
struct SetStmt : public TreeNode {

    SetKnobType set_knob_type_;
    std::string output_file_;
    bool bool_val_;

    SetStmt(SetKnobType &type, bool bool_value) : 
        set_knob_type_(type), bool_val_(bool_value) { }

    SetStmt(std::string &output_file, bool bool_value) :
            output_file_(output_file), bool_val_(bool_value) { }
};

struct LoadStmt : public TreeNode{
    std::string file_name;
    std::string tab_name;

    LoadStmt(std::string file_name_, std::string tab_name_) :
            file_name(std::move(file_name_)), tab_name(std::move(tab_name_)) {}
};

struct SetHelp : public TreeNode{
    std::string output_file;
    bool is_on;
    SetHelp(std::string &output_file_,bool is_on_):
        output_file(std::move(output_file_)),is_on(is_on_){}
};

// Semantic value
struct SemValue {
    int sv_int;
    float sv_float;
    std::string sv_str;
    bool sv_bool;
    OrderByDir sv_orderby_dir;
    std::vector<std::string> sv_strs;

    std::shared_ptr<TreeNode> sv_node;

    SvCompOp sv_comp_op;

    SvAgreType sv_type_agre;

    std::shared_ptr<TypeLen> sv_type_len;

    std::shared_ptr<Field> sv_field;
    std::vector<std::shared_ptr<Field>> sv_fields;

    std::shared_ptr<Expr> sv_expr;

    std::shared_ptr<Value> sv_val;
    std::vector<std::shared_ptr<Value>> sv_vals;

    std::shared_ptr<Col> sv_col;
    std::vector<std::shared_ptr<Col>> sv_cols;

    std::shared_ptr<Col> sv_agrecol;

    std::shared_ptr<SetClause> sv_set_clause;
    std::vector<std::shared_ptr<SetClause>> sv_set_clauses;

    std::shared_ptr<BinaryExpr> sv_cond;
    std::vector<std::shared_ptr<BinaryExpr>> sv_conds;

    std::shared_ptr<OrderBy> sv_orderby;

    std::shared_ptr<Col> sv_group_by_col;
    std::vector<std::shared_ptr<Col>> sv_group_by_cols;

    std::shared_ptr<BinaryExpr> sv_having_cond;
    std::vector<std::shared_ptr<BinaryExpr>> sv_having_conds;

    SetKnobType sv_setKnobType;

    SvUpdateOp sv_update_op;
};

extern std::shared_ptr<ast::TreeNode> parse_tree;

}

#define YYSTYPE ast::SemValue
