/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "log_recovery.h"

/**
 * @description: analyze阶段，需要获得脏页表（DPT）和未完成的事务列表（ATT）
 *
 */

void RecoveryManager::analyze() {
    // 开辟一个用于从磁盘读取log到内存的一个空间
    // 遍历到的buffer的下标
    int buffer_offset = 0;
    // 遍历到的磁盘上log文件的下标
    int file_offset = 0;
    // 每次读取的字节数
    int read_bytes;
    txn_id_t max_id = -1;
    while((read_bytes = disk_manager_->read_log(buffer_.buffer_,LOG_BUFFER_SIZE,file_offset)) > 0){
        buffer_.offset_ = read_bytes - 1;
        //获取一个缓冲区的部分，看成获取一个log页
        //将页里的log逐一取出分析
        // log共有部分 log_type,lsn,log_tot_len_,log_tid,prev_lsn
        int log_start_offset = 0;
        while(log_start_offset <= buffer_.offset_ ){
            if((log_start_offset + OFFSET_LOG_TOT_LEN + sizeof(uint32_t)) > LOG_BUFFER_SIZE)break;
            uint32_t total_len = *reinterpret_cast<const uint32_t *>(buffer_.buffer_ + log_start_offset + OFFSET_LOG_TOT_LEN);
            //assert(total_len);
            //if(total_len == 0 || total_len + log_start_offset > LOG_BUFFER_SIZE)break;
            if(total_len + log_start_offset > LOG_BUFFER_SIZE)break;
            auto log_type = *reinterpret_cast<const LogType *>(buffer_.buffer_ + log_start_offset + OFFSET_LOG_TYPE);
            auto rec = generator_type2log(log_type);
            rec->deserialize(buffer_.buffer_ + log_start_offset);//这里仅仅获取抽象类内的五个原始数据？

            //更新map
            lsn_map_[rec->lsn_] = std::make_pair(file_offset,total_len);
            read_list.emplace_back(rec->lsn_);
            undo_txn_map_[rec->log_tid_] = rec->lsn_;//目的：将所有lsn标记为undo，后续移除

            if(rec->log_type_ == LogType::commit){
                undo_txn_map_.erase(rec->log_tid_);
                redo_txn_map_[rec->log_tid_] = rec->lsn_;
            }else if(rec->log_type_ == LogType::ABORT) {
                undo_txn_map_.erase(rec->log_tid_);
                redo_txn_map_[rec->log_tid_] = rec->lsn_;
            }
            else if(rec->log_type_ == LogType::begin){
                    max_id = rec->log_tid_;
            }else{
                //dml_update_dpt_map(rec);
                redo_list.emplace_back(rec->lsn_);
            }
            log_start_offset += (int)total_len;
            file_offset += (int)total_len;
        }
    }
    this->log_manager_->setGlobalLsn(static_cast<lsn_t>(lsn_map_.size()));
    this->log_manager_->setPersistLsn(static_cast<lsn_t>(lsn_map_.size() - 1));
    this->log_manager_->setLastTid(max_id);
}

/**
 * @description: 重做所有未落盘的操作
 * 以页为单位操作
 */
//void RecoveryManager::redo() {
//    //从ddp重做，其中对每个页已经维护了有序的lsn操作
//    for(const auto& [page_id,redo_logs_in_page] : dpt_map_) {
//        auto rm_file_handle = redo_logs_in_page.table_file_;
//        for(const auto lsn : redo_logs_in_page.redo_logs_){
//            //根据lsn找到具体的日志
//            auto log_rec = getRecByLsn(lsn);
//            //assert(log_rec->log_type_ != LogType::commit);
//            switch(log_rec->log_type_){
//                case LogType::INSERT:
//                    redoInsertLog(dynamic_cast<InsertLogRecord *>(log_rec),page_id,rm_file_handle);
//                    break;
//                case LogType::DELETE:
//                    redoDeleteLog(dynamic_cast<DeleteLogRecord *>(log_rec),page_id,rm_file_handle);
//                    break;
//                case LogType::UPDATE:
//                    redoUpdateLog(dynamic_cast<UpdateLogRecord *>(log_rec),page_id,rm_file_handle);
//                    break;
//                case LogType::ABORT:
//                case LogType::commit:
//                    log_rec->format_print();
//                    break;
//                default:
//                    //assert(0);
//                    break;
//            }
//        }
//    }
//}

void RecoveryManager::redo() {
    //从ddp重做，其中对每个页已经维护了有序的lsn操作
    for(const auto lsn : read_list) {
        auto log_rec = getRecByLsn(lsn);
        //assert(log_rec->log_type_ != LogType::commit);
        RmFileHandle *file_handle;
        //RmPageHandle *page_handle;
        switch(log_rec->log_type_) {
            case LogType::INSERT: {
                auto insert_rec = dynamic_cast<InsertLogRecord *>(log_rec);
                //redoInsertLog(dynamic_cast<InsertLogRecord *>(log_rec),page_id,rm_file_handle);
                file_handle = sm_manager_->fhs_[std::string(insert_rec->table_name_,
                                                            insert_rec->table_name_size_)].get();
                auto page_handle = file_handle->fetch_page_handle(insert_rec->rid_.page_no);
                if (page_handle.page->get_rec_lsn() >= insert_rec->lsn_) {
                    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), false);
                    break;
                }
                //file_handle->insert_record(insert_rec->insert_value_.data, nullptr);
                allocPage(insert_rec->rid_,file_handle);
                file_handle->insert_record(insert_rec->rid_,insert_rec->insert_value_.data);//TODO
                page_handle.page->set_page_lsn(insert_rec->lsn_);
                buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
                break;
            }
            case LogType::DELETE: {
                auto del_rec = dynamic_cast<DeleteLogRecord *>(log_rec);

                file_handle = sm_manager_->fhs_[std::string(del_rec->table_name_, del_rec->table_name_size_)].get();
                auto page_handle = file_handle->fetch_page_handle(del_rec->rid_.page_no);
                if (page_handle.page->get_rec_lsn() >= del_rec->lsn_) {
                    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), false);
                    break;
                }
                file_handle->delete_record(del_rec->rid_, nullptr);
                page_handle.page->set_page_lsn(del_rec->lsn_);
                buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
                //redoDeleteLog(dynamic_cast<DeleteLogRecord *>(log_rec),page_id,rm_file_handle);
                break;
            }
            case LogType::UPDATE: {
                auto update_rec = dynamic_cast<UpdateLogRecord *>(log_rec);
                file_handle = sm_manager_->fhs_[std::string(update_rec->table_name_,
                                                            update_rec->table_name_size_)].get();
                auto page_handle = file_handle->fetch_page_handle(update_rec->rid_.page_no);
                if (page_handle.page->get_rec_lsn() >= update_rec->lsn_) {
                    buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), false);
                    break;
                }
                file_handle->update_record(update_rec->rid_, update_rec->new_value_.data, nullptr);
                page_handle.page->set_page_lsn(update_rec->lsn_);
                buffer_pool_manager_->unpin_page(page_handle.page->get_page_id(), true);
                break;
                //redoUpdateLog(dynamic_cast<UpdateLogRecord *>(log_rec),page_id,rm_file_handle);
            }

//            case LogType::ABORT:
//            case LogType::commit:
//                log_rec->format_print();
//                break;
            default:
                //assert(0);
                break;
        }
        //log_rec->format_print();
    }
}

/**
 * @description: 回滚未完成的事务
 *
 */
void RecoveryManager::undo() {
    //TODO 实现循环逻辑

    for(auto it = read_list.rbegin(); it != read_list.rend(); ++it){//这里获取第一个需要重做的日志
        //对于第一个，开始重做
        if(*it == INVALID_LSN)continue;
        LogRecord* undo_log = getRecByLsn(*it);
        if(undo_txn_map_.find(undo_log->log_tid_) == undo_txn_map_.end())continue;
        switch(undo_log->log_type_){
            case LogType::INSERT:{
                undoInsertLog(dynamic_cast<InsertLogRecord *>(undo_log));
                break;
            }
            case LogType::DELETE:{
                undoDeleteLog(dynamic_cast<DeleteLogRecord *>(undo_log));
                break;
            }
            case LogType::UPDATE:{
                undoUpdateLog(dynamic_cast<UpdateLogRecord *>(undo_log));
                break;
            }
            case LogType::begin:{
                auto begin_log = dynamic_cast<BeginLogRecord *>(undo_log);
                auto abort_log = AbortLogRecord(begin_log->log_tid_,undo_txn_map_[begin_log->log_tid_]);
                log_manager_->add_log_to_buffer(&abort_log);
                //undo_txn_map_
                break;
            }
            default:
                //可能begin
                break;
        }
    }

    //sorted_vector.clear();
//    for(auto [tid,lsn] : undo_txn_map_){
//        auto abortlog = AbortLogRecord(tid, lsn);
//        log_manager_->add_log_to_buffer(&abortlog);
//    }
    undo_txn_map_.clear();
    //index索引
    maintain_index_consistency();
}
//void RecoveryManager::undo() {
//    //TODO 实现循环逻辑
//    //undo只需要每一个被修改的
////    std::vector<std::pair<txn_id_t, lsn_t>> sorted_vector;
////
////    // 将 undo_txn_map_ 中的键值对插入到 vector 中
////    for (const auto& [undo_txn_id, undo_lsn] : undo_txn_map_) {
////        sorted_vector.emplace_back(undo_txn_id, undo_lsn);
////    }
////    // 使用 std::sort 对 vector 进行排序，依据为 undo_lsn 从大到小
////    std::sort(sorted_vector.begin(), sorted_vector.end(), [](const std::pair<txn_id_t, lsn_t>& a, const std::pair<txn_id_t, lsn_t>& b) {
////        return a.second > b.second;
////    });
//
//    for(auto it = redo_list.rbegin(); it != redo_list.rend(); ++it){//这里获取第一个需要重做的日志
//        //对于第一个，开始重做
//        LogRecord* undo_log = getRecByLsn(*it);
//        if(redo_txn_map_.find(undo_log->log_tid_) != redo_txn_map_.end() || *it == INVALID_LSN)continue;
//        lsn_t prev_lsn = *it;
//        while(prev_lsn != INVALID_LSN){
//            //assert(undo_log->log_type_ != LogType::commit);
//            switch(undo_log->log_type_){
//                case LogType::INSERT:
//                    undoInsertLog(dynamic_cast<InsertLogRecord *>(undo_log));
//                    break;
//                case LogType::DELETE:
//                    undoDeleteLog(dynamic_cast<DeleteLogRecord *>(undo_log));
//                    break;
//                case LogType::UPDATE:
//                    undoUpdateLog(dynamic_cast<UpdateLogRecord *>(undo_log));
//                    break;
//                default:
//                    //可能begin
//                    break;
//            }
//            prev_lsn = undo_log->prev_lsn_;
//            undo_log = getRecByLsn(prev_lsn);
//        }
//    }
//    //sorted_vector.clear();
//    undo_txn_map_.clear();
//    //index索引
//    maintain_index_consistency();
//}
/**
 * @description:记录dpt
 * @param rec
 */
void RecoveryManager::dml_update_dpt_map(LogRecord *rec) {
    auto process_record = [this](const std::string &table_name, const Rid &rid, lsn_t lsn) {
        auto rm_file_hdl = sm_manager_->fhs_[table_name].get();
        int fd = rm_file_hdl->GetFd();
        auto page_id = PageId(fd, rid.page_no);
        Page *page = sm_manager_->get_bpm()->fetch_page(page_id);
        if(page->get_page_lsn() >= lsn){
            sm_manager_->get_bpm()->unpin_page(page_id,false);
            return ;
        }
        sm_manager_->get_bpm()->unpin_page(page_id,false);

        if (dpt_map_.find(page_id) == dpt_map_.end()) {
            dpt_map_[page_id] = RedoLogsInPage(rm_file_hdl);
        }
        dpt_map_[page_id].redo_logs_.emplace_back(lsn);
    };

    switch (rec->log_type_) {
        case LogType::DELETE: {
            auto del_rec = reinterpret_cast<DeleteLogRecord *>(rec);
            process_record(std::string(del_rec->table_name_, del_rec->table_name_size_), del_rec->rid_, rec->lsn_);
            break;
        }
        case LogType::INSERT: {
            auto insert_rec = reinterpret_cast<InsertLogRecord *>(rec);
            process_record(std::string(insert_rec->table_name_, insert_rec->table_name_size_), insert_rec->rid_, rec->lsn_);
            break;
        }
        case LogType::UPDATE: {
            auto update_rec = reinterpret_cast<UpdateLogRecord *>(rec);
            process_record(std::string(update_rec->table_name_, update_rec->table_name_size_), update_rec->rid_, rec->lsn_);
            break;
        }
        default:
            assert(0);
    }
}

void RecoveryManager::redoInsertLog(InsertLogRecord *rec,PageId page_id,RmFileHandle* file_handle) {
    //if(redo_txn_map_.find(rec->log_tid_) == redo_txn_map_.end())return;
    auto page = buffer_pool_manager_->fetch_page(page_id);
    //开始修改
    if(page->get_page_lsn() >= rec->lsn_){
        buffer_pool_manager_->unpin_page(page->get_page_id(), false);
        return;
    }
    //assert(page->get_page_lsn() < rec->lsn_);
    //TODO 插入固定位置
    file_handle->insert_record(rec->insert_value_.data, nullptr);
    page->set_page_lsn(rec->lsn_);
    buffer_pool_manager_->unpin_page(page->get_page_id(),true);
    rec->format_print();
}

void RecoveryManager::redoDeleteLog(DeleteLogRecord *rec,PageId page_id,RmFileHandle* file_handle) {
    //if(redo_txn_map_.find(rec->log_tid_) == redo_txn_map_.end())return;
    auto page = buffer_pool_manager_->fetch_page(page_id);

    //修改已经被提交到内存中
    if(page->get_page_lsn() >= rec->lsn_){
        buffer_pool_manager_->unpin_page(page->get_page_id(), false);
        return;
    }
    //assert(page->get_page_lsn() < rec->lsn_);
    file_handle->delete_record(rec->rid_, nullptr);//单线程
    page->set_page_lsn(rec->lsn_);
    buffer_pool_manager_->unpin_page(page->get_page_id(),true);
    //rec->format_print();
}

void RecoveryManager::redoUpdateLog(UpdateLogRecord *rec,PageId page_id,RmFileHandle* file_handle) {
    //if(redo_txn_map_.find(rec->log_tid_) == redo_txn_map_.end())return;
    auto page = buffer_pool_manager_->fetch_page(page_id);

    if(page->get_page_lsn() >= rec->lsn_){
        buffer_pool_manager_->unpin_page(page->get_page_id(), false);
        return;
    }
    //assert(page->get_page_lsn() < rec->lsn_);
    file_handle->update_record(rec->rid_,rec->new_value_.data, nullptr);
    page->set_page_lsn(rec->lsn_);
    buffer_pool_manager_->unpin_page(page->get_page_id(),true);
    rec->format_print();
}

/**
 * @brief 确保rid插入后，file handle 有足够的页
 * @param rid 插入的数据，在故障前的位置
 * @param fh_
 */
void RecoveryManager::allocPage(Rid& rid, RmFileHandle* fh_){
    while(rid.page_no >= fh_->get_file_hdr().num_pages){
        auto rm_page_hd = fh_->create_new_page_handle();
        sm_manager_->get_bpm()->unpin_page(rm_page_hd.page->get_page_id(), false);
    }
}

/**
 *
 * @param rec
 * @brief 在缓冲池内完成undo操作，并且记录undo日志
 */
void RecoveryManager::undoInsertLog(InsertLogRecord *rec) {
    auto rid = rec->rid_;
    std::string table_name(rec->table_name_,rec->table_name_size_);
    auto file_handle = sm_manager_->fhs_[table_name].get();
    auto fd = sm_manager_->fhs_[table_name]->GetFd();
    auto page = buffer_pool_manager_->fetch_page({fd,rid.page_no});
    //auto data = page->get_data();
    //auto page_lsn = page->get_page_lsn();
    if(page->get_rec_lsn() > rec->lsn_){
        buffer_pool_manager_->unpin_page(page->get_page_id(),false);
        return;
    }
    //undo时也需要日志记录
    auto log = DeleteLogRecord(rec->log_tid_,undo_txn_map_[rec->log_tid_],rec->insert_value_,rid,table_name);

    auto lsn = log_manager_->add_log_to_buffer(&log);
    file_handle->delete_record(rid, nullptr);
    page->set_page_lsn(lsn);
    undo_txn_map_[rec->log_tid_] = lsn;
    buffer_pool_manager_->unpin_page(page->get_page_id(),true);

    //rec->format_print();
}

void RecoveryManager::undoDeleteLog(DeleteLogRecord *rec) {
    auto rid = rec->rid_;
    std::string table_name(rec->table_name_,rec->table_name_size_);
    auto file_handle = sm_manager_->fhs_[table_name].get();
    auto fd = sm_manager_->fhs_[table_name]->GetFd();
    auto page = buffer_pool_manager_->fetch_page({fd,rid.page_no});

    if(page->get_rec_lsn() > rec->lsn_){
        buffer_pool_manager_->unpin_page(page->get_page_id(),false);
        return;
    }
    //undo时也需要日志记录
    auto log = InsertLogRecord(rec->log_tid_,undo_txn_map_[rec->log_tid_],rec->delete_value_,rid,table_name);

    auto lsn = log_manager_->add_log_to_buffer(&log);
    file_handle->insert_record(rid,rec->delete_value_.data);
    page->set_page_lsn(lsn);
    undo_txn_map_[rec->log_tid_] = lsn;
    buffer_pool_manager_->unpin_page(page->get_page_id(),true);

    //rec->format_print();
}

void RecoveryManager::undoUpdateLog(UpdateLogRecord *rec) {
    auto rid = rec->rid_;
    std::string table_name(rec->table_name_,rec->table_name_size_);
    auto file_handle = sm_manager_->fhs_[table_name].get();
    auto fd = sm_manager_->fhs_[table_name]->GetFd();
    auto page = buffer_pool_manager_->fetch_page({fd,rid.page_no});

    if(page->get_rec_lsn() > rec->lsn_){
        buffer_pool_manager_->unpin_page(page->get_page_id(),false);
        return;
    }

    //undo时也需要日志记录
    auto log = UpdateLogRecord(rec->log_tid_,undo_txn_map_[rec->log_tid_],rec->new_value_,rec->old_value_,rid,table_name);

    auto lsn = log_manager_->add_log_to_buffer(&log);
    file_handle->update_record(rid,rec->old_value_.data,nullptr);
    page->set_page_lsn(lsn);
    undo_txn_map_[rec->log_tid_] = lsn;
    buffer_pool_manager_->unpin_page(page->get_page_id(),true);
    //rec->format_print();
}

/**
 * @brief:保持index内的数据一致性
 */
void RecoveryManager::maintain_index_consistency() {
    for(auto iter = this->sm_manager_->db_.getTabs().begin(); iter != this->sm_manager_->db_.getTabs().end(); ++iter){
        //当前表内所有索引
        auto table_name = iter->first;
        std::vector<IndexMeta> tmp_indexes;
        for(auto single_index_iter = iter->second.indexes.begin(); single_index_iter != iter->second.indexes.end(); single_index_iter++){
            tmp_indexes.push_back(*single_index_iter);
        }

        for(auto single_index_meta : tmp_indexes){
            //一个索引
            //先获得句柄文件
            std::vector<std::string> col_names;
            for(const auto& col_meta : single_index_meta.cols){
                col_names.push_back(col_meta.name);
            }
            sm_manager_->recovery_create_index(table_name,col_names);
        }
    }
}

